@@ -1,3 +1,4 @@
+/*  LAST EDIT: Mon Oct  2 20:28:55 1995 by Thorsten Kukuk (kukuk)  */
 /*
 ** server.c			YP server routines.
 **
@@ -21,6 +22,10 @@
 ** Cambridge, MA 02139, USA.
 **
 ** Author: Peter Eriksson <pen@signum.se>
+**
+** Hacked some features from the FreeBSD Port and from the yps-0.21
+** package back in the original code and fixed some bugs:
+** Thorsten Kukuk <kukuk@uni-paderborn.de>
 */
 
 #include "system.h"
@@ -37,26 +42,51 @@
 #include <sys/socket.h>
 #include <arpa/inet.h>
 #include <netdb.h>
-
-#if TCP_WRAPPER
 #include <syslog.h>
-#include "log_tcp.h"
+#include <sys/wait.h>
+#include <sys/param.h>
+#include <errno.h>
+#include <stdarg.h>
+
+#if HAVE_LIBWRAP
+#include <tcpd.h>
 int allow_severity=LOG_INFO;
 int deny_severity=LOG_WARNING;
 #endif
 
+#ifndef HAVE__DNS_GETHOSTBYNAME
+#define dns_gethostbyname gethostbyname
+#define dns_gethostbyaddr gethostbyaddr
+#else
+#define dns_gethostbyname _dns_gethostbyname
+#define dns_gethostbyaddr _dns_gethostbyaddr
+#endif
+
+
 extern char *progname;
 
 int debug_flag = 0;
 int dns_flag   = 0;
+int children   = 0;
+int forked     = 0;
+
+void Perror(char *fmt, ...)
+{
+   va_list ap;
+   va_start(ap, fmt);
+   if (debug_flag)
+     vfprintf(stderr, fmt, ap);
+   else
+     vsyslog(LOG_NOTICE, fmt, ap);
+   va_end(ap);
+}   
 
 static void Gperror(const char *str)
 {
-    fprintf(stderr, "%s: GDBM Error Code #%d\n", str, gdbm_errno);
+    Perror("%s: GDBM Error Code #%d\n", str, gdbm_errno);
 }
 
-
-#ifdef NEED_GETHOSTNAME
+#ifndef HAVE_GETHOSTNAME
 #include <sys/utsname.h>
 
 int gethostname(char *buf, int buflen)
@@ -88,7 +118,7 @@ int gethostname(char *buf, int buflen)
 */
 static int is_valid_host(struct sockaddr_in *sin)
 {
-#if TCP_WRAPPER
+#if HAVE_LIBWRAP
     extern int hosts_ctl(char *, char *, char *, char *);
     int status;
     static long oldaddr=0;	/* so we dont log multiple times */
@@ -104,12 +134,16 @@ static int is_valid_host(struct sockaddr_in *sin)
     h = (hp && hp->h_name) ? hp->h_name : NULL;
 #endif
 
+#ifndef FROM_UNKNOWN
+#define FROM_UNKNOWN STRING_UNKNOWN
+#endif
+
     status = hosts_ctl(progname,
 		       h?h:FROM_UNKNOWN,
 		       inet_ntoa(sin->sin_addr),
 		       "");
 
-    if (sin->sin_addr.s_addr != oldaddr || status != oldstatus ) {
+    if (!status && (sin->sin_addr.s_addr != oldaddr || status != oldstatus )) {
 	syslog(status?allow_severity:deny_severity,
 	       "%sconnect from %s\n",status?"":"refused ",
 	       h?h:inet_ntoa(sin->sin_addr));
@@ -135,7 +169,7 @@ void *ypproc_null_2_svc(void *dummy,
 	return NULL;
 
     if (debug_flag)
-	fprintf(stderr, "ypproc_null() [From: %s:%d]\n",
+	Perror("ypproc_null() [From: %s:%d]\n",
 		inet_ntoa(rqhost->sin_addr),
 		ntohs(rqhost->sin_port));
 
@@ -176,7 +210,7 @@ bool_t *ypproc_domain_2_svc(domainname *name,
     rqhost = svc_getcaller(rqstp->rq_xprt);
 	
     if (debug_flag)
-	fprintf(stderr, "ypproc_domain(\"%s\") [From: %s:%d]\n",
+	Perror("ypproc_domain(\"%s\") [From: %s:%d]\n",
 		*name,
 		inet_ntoa(rqhost->sin_addr),
 		ntohs(rqhost->sin_port));
@@ -184,7 +218,7 @@ bool_t *ypproc_domain_2_svc(domainname *name,
     if (!is_valid_host(rqhost))
     {
 	if (debug_flag)
-	    fprintf(stderr, "\t-> Ignored (not a valid source host)\n");
+	    Perror("\t-> Ignored (not a valid source host)\n");
 	
 	return NULL;
     }
@@ -195,7 +229,7 @@ bool_t *ypproc_domain_2_svc(domainname *name,
 	result = FALSE;
     
     if (debug_flag)
-	fprintf(stderr, "\t-> %s.\n",
+	Perror("\t-> %s.\n",
 		(result == TRUE ? "Ok" : "Not served by us"));
     
     return &result;
@@ -212,7 +246,7 @@ bool_t *ypproc_domain_nonack_2_svc(domainname *name,
     rqhost = svc_getcaller(rqstp->rq_xprt);
 
     if (debug_flag)
-	fprintf(stderr, "ypproc_domain_nonack(\"%s\") [From: %s:%d]\n",
+	Perror("ypproc_domain_nonack(\"%s\") [From: %s:%d]\n",
 		*name,
 		inet_ntoa(rqhost->sin_addr),
 		ntohs(rqhost->sin_port));
@@ -220,7 +254,7 @@ bool_t *ypproc_domain_nonack_2_svc(domainname *name,
     if (!is_valid_host(rqhost))
     {
 	if (debug_flag)
-	    fprintf(stderr, "\t-> Ignored (not a valid source host)\n");
+	    Perror("\t-> Ignored (not a valid source host)\n");
 	
 	return NULL;
     }
@@ -228,14 +262,14 @@ bool_t *ypproc_domain_nonack_2_svc(domainname *name,
     if (!is_valid_domain(*name))
     {
 	if (debug_flag)
-	    fprintf(stderr, "\t-> Ignored (not a valid domain)\n");
+	    Perror("\t-> Ignored (not a valid domain)\n");
 
 	/* Bail out and don't return any RPC value */
 	return NULL;
     }
     
     if (debug_flag)
-	fprintf(stderr, "\t-> OK.\n");
+	Perror("\t-> OK.\n");
     
     result = TRUE;
     return &result;
@@ -253,27 +287,25 @@ static GDBM_FILE open_database(const char *domain,
 
 
     if (debug_flag)
-	fprintf(stderr, "\topen_database(\"%s\", \"%s\")\n",
+	Perror("\topen_database(\"%s\", \"%s\")\n",
 		domain, map);
     
     if (map[0] == '.' || strchr(map, '/'))
     {
 	if (debug_flag)
-	    fprintf(stderr, "\t\t->Returning 0\n");
+	    Perror("\t\t->Returning 0\n");
 	
 	return 0;
     }
 
-    strcpy(buf, domain);
-    strcat(buf, "/");
-    strcat(buf, map);
+    sprintf(buf, "%s/%s", domain, map);
     
-    dbp = gdbm_open(buf, 0, O_RDONLY, 0, NULL);
+    dbp = gdbm_open(buf, 0, GDBM_READER, 0, NULL);
 
     if (debug_flag && dbp == NULL)
 	Gperror("gdbm_open()");
     else if (debug_flag)
-	fprintf(stderr, "\t\t->Returning OK!\n");
+	Perror("\t\t->Returning OK!\n");
     
     return dbp;
 }
@@ -298,7 +330,7 @@ int read_database(GDBM_FILE dbp,
     if (ikey == NULL || ikey->dptr == NULL)
     {
 	if (debug_flag)
-	    fprintf(stderr, "\tread_database(), gdbm_firstkey()\n");
+	    Perror("\tread_database(), gdbm_firstkey()\n");
 	
 	ckey = gdbm_firstkey(dbp);
 	first_flag = 1;
@@ -306,7 +338,7 @@ int read_database(GDBM_FILE dbp,
     else
     {
 	if (debug_flag)
-	    fprintf(stderr, "\tread_database(), gdbm_nextkey()\n");
+	    Perror("\tread_database(), gdbm_nextkey()\n");
 	
 	if ((flags & F_NEXT))
 	    ckey = gdbm_nextkey(dbp, *ikey);
@@ -338,9 +370,9 @@ int read_database(GDBM_FILE dbp,
 		else
 		    return YP_FALSE;
 	}
-	
-	if ((flags & F_ALL) || strncmp(dval->dptr, "YP_", 3) != 0)
-	{
+
+	if ((flags & F_ALL) || strncmp(ckey.dptr, "YP_", 3) != 0)
+	{		      
 	    if (okey)
 		*okey = ckey;
 	    else if (ikey == NULL || ikey->dptr != ckey.dptr)
@@ -358,7 +390,7 @@ int read_database(GDBM_FILE dbp,
 	if (ikey == NULL || ckey.dptr != ikey->dptr)
 	    free(ckey.dptr);
 
-	if (ckey.dptr == NULL)
+	if (ckey.dptr == NULL || nkey.dptr == NULL)
 	    return YP_NOMORE;
 
 	ckey = nkey;
@@ -376,9 +408,7 @@ static unsigned long get_dtm(const char *domain,
     char buf[1025];
 
     
-    strcpy(buf, domain);
-    strcat(buf, "/");
-    strcat(buf, map);
+    sprintf(buf, "%s/%s" ,domain, map);
     
     if (stat(buf, &sbuf) < 0)
 	return 0;
@@ -401,15 +431,15 @@ ypresp_val *ypproc_match_2_svc(ypreq_key *key,
 
     if (debug_flag)
     {
-	fprintf(stderr, "ypproc_match(): [From: %s:%d]\n",
+	Perror("ypproc_match(): [From: %s:%d]\n",
 		inet_ntoa(rqhost->sin_addr),
 		ntohs(rqhost->sin_port));
 		
-	fprintf(stderr, "\t\tdomainname = \"%s\"\n",
+	Perror("\t\tdomainname = \"%s\"\n",
 		key->domain);
-	fprintf(stderr, "\t\tmapname = \"%s\"\n",
+	Perror("\t\tmapname = \"%s\"\n",
 		key->map);
-	fprintf(stderr, "\t\tkeydat = \"%.*s\"\n",
+	Perror("\t\tkeydat = \"%.*s\"\n",
 		(int) key->key.keydat_len,
 		key->key.keydat_val);
     }
@@ -417,7 +447,7 @@ ypresp_val *ypproc_match_2_svc(ypreq_key *key,
     if (!is_valid_host(rqhost))
     {
 	if (debug_flag)
-	    fprintf(stderr, "\t-> Ignored (not a valid source host)\n");
+	    Perror("\t-> Ignored (not a valid source host)\n");
 	
 	return NULL;
     }
@@ -429,7 +459,7 @@ ypresp_val *ypproc_match_2_svc(ypreq_key *key,
 	result.val.valdat_val = NULL;
     }
     
-    if (key->domain[0] == '\0' || key->map[0] == '\0')
+    if (key->domain[0] == '\0' || key->map[0] == '\0' || key->key.keydat_len == 0 || key->key.keydat_val[0] == '\0')
 	result.stat = YP_BADARGS;
     else if (!is_valid_domain(key->domain))
 	result.stat = YP_NODOM;
@@ -460,11 +490,11 @@ ypresp_val *ypproc_match_2_svc(ypreq_key *key,
 	if (debug_flag)
 	{
 	if (result.stat == YP_TRUE)
-	    fprintf(stderr, "\t-> Value = \"%.*s\"\n",
+	    Perror("\t-> Value = \"%.*s\"\n",
 		    (int) result.val.valdat_len,
 		    result.val.valdat_val);
 	else
-	    fprintf(stderr, "\t-> Error #%d\n", result.stat);
+	    Perror("\t-> Error #%d\n", result.stat);
     }
 
 
@@ -472,67 +502,93 @@ ypresp_val *ypproc_match_2_svc(ypreq_key *key,
     ** Do the jive thing if we didn't find the host in the YP map
     ** and we have enabled the magic DNS lookup stuff.
     **
-    ** XXX Perhaps this should be done in a sub-process for performance
-    **     reasons. Later.
+    ** DNS lookups are handled in a subprocess so that the server
+    ** doesn't block while waiting for requests to complete.
     */
-    if (result.stat != YP_TRUE && dns_flag)
+    if (result.stat != YP_TRUE && strstr(key->map, "hosts") && dns_flag)
     {
-	struct hostent *he;
+        struct hostent *he;
+        if (children < MAX_CHILDREN && fork())
+        {
+           children++;
+           return NULL;
+        }
+        else
+           forked++;
+           
 	key->key.keydat_val[key->key.keydat_len] = '\0';
 
 
 	if (debug_flag)
-	    fprintf(stderr, "Doing DNS lookup of %s\n", key->key.keydat_val);
+	    Perror("Doing DNS lookup of %s\n", key->key.keydat_val);
 	
 	if (strcmp(key->map, "hosts.byname") == 0)
-	    he = gethostbyname(key->key.keydat_val);
+	    he = dns_gethostbyname(key->key.keydat_val);
 	else  if (strcmp(key->map, "hosts.byaddr") == 0)
-	{
-	    long a;
+	  {
+	      long a;
 
-	    a = inet_addr(key->key.keydat_val);
-	    he = gethostbyaddr((const char *) &a, sizeof(long), AF_INET);
-	}
+	      a = inet_addr(key->key.keydat_val);
+	      he = dns_gethostbyaddr((const char *) &a, sizeof(long), AF_INET);
+	  }
 	else
 	    he = NULL;
-	    
+
 	if (he)
 	{
-	    int len, i;
-	    char *cp;
-	    
-	    
-	    if (debug_flag)
-		fprintf(stderr, "\t-> OK (%s/%s)\n",
-			he->h_name,
-			inet_ntoa(* (struct in_addr *) he->h_addr));
+	  int len, i, j;
+	  char *cp;
 
-	    len = 16; /* Numerical IP address */
-	    len += strlen(he->h_name) + 1;
-	    for (i = 0; he->h_aliases[i]; i++)
-		len += strlen(he->h_aliases[i]) + 1;
-	    
+	  if (debug_flag)
+		Perror("\t-> OK (%s/%s)\n",
+		       he->h_name,
+		       inet_ntoa(* (struct in_addr *) he->h_addr));
 
-	    cp = malloc(len+1);
-	    strcpy(cp, inet_ntoa(* (struct in_addr *) he->h_addr));
-	    strcat(cp, " ");
-	    strcat(cp, he->h_name);
+	  for (j = 0; he->h_addr_list[j]; j++)
+	      ;
+	  
+	  len = 16; /* Numerical IP address */
+	  len += strlen(he->h_name) + 1;
+	  for (i = 0; he->h_aliases[i]; i++)
+	      len += strlen(he->h_aliases[i]) +1;
 
-	    for (i = 0; he->h_aliases[i]; i++)
+	  len *= j;
+	  cp = malloc(len+1);
+	  
+	  strcpy(cp, inet_ntoa(* (struct in_addr *) he->h_addr_list[0]));
+	  strcat(cp, "	");
+	  strcat(cp, he->h_name);
+
+	  for(i = 0; he->h_aliases[i]; i++)
 	    {
-		strcat(cp, " ");
-		strcat(cp, he->h_aliases[i]);
+	      strcat(cp, " ");
+	      strcat(cp, he->h_aliases[i]);
 	    }
 
-	    result.val.valdat_len = len;
-	    result.val.valdat_val = cp;
-	    result.stat = YP_TRUE;
-	}
+	  for (j = 1; he->h_addr_list[j]; j++) {
+	      strcat(cp, "\n");
+	      strcat(cp, inet_ntoa(* (struct in_addr *) he->h_addr_list[j]));
+	      strcat(cp, "	");
+	      strcat(cp, he->h_name);
+
+	      for (i = 0; he->h_aliases[i]; i++)
+	      {
+		  strcat(cp, " ");
+		  strcat(cp, he->h_aliases[i]);
+	      }
+	  }
+	  strcat(cp, "\n");
+	  len = strlen(cp);
+	  
+	  result.val.valdat_len = len;
+	  result.val.valdat_val = cp;
+	  result.stat = YP_TRUE;
+        }
 	else
 	{
 	    if (debug_flag)
-		fprintf(stderr, "\t-> Not Found\n");
-
+		Perror("\t-> Not Found\n");
+	    
 	    result.stat = YP_NOKEY;
 	}
     }
@@ -547,20 +603,20 @@ ypresp_key_val *ypproc_first_2_svc(ypreq_nokey *key,
 {
     static ypresp_key_val result;
     struct sockaddr_in *rqhost;
-    
+ 
 
     rqhost = svc_getcaller(rqstp->rq_xprt);
 
     if (debug_flag)
     {
-	fprintf(stderr, "ypproc_first(): [From: %s:%d]\n",
+	Perror("ypproc_first(): [From: %s:%d]\n",
 		inet_ntoa(rqhost->sin_addr),
 		ntohs(rqhost->sin_port));
 		
-	fprintf(stderr, "\tdomainname = \"%s\"\n", key->domain);
-	fprintf(stderr, "\tmapname = \"%s\"\n", key->map);
+	Perror("\tdomainname = \"%s\"\n", key->domain);
+	Perror("\tmapname = \"%s\"\n", key->map);
 #if 0
-	fprintf(stderr, "\tkeydat = \"%.*s\"\n",
+	Perror("\tkeydat = \"%.*s\"\n",
 		(int) key->key.keydat_len,
 		key->key.keydat_val);
 #endif
@@ -569,7 +625,7 @@ ypresp_key_val *ypproc_first_2_svc(ypreq_nokey *key,
     if (!is_valid_host(rqhost))
     {
 	if (debug_flag)
-	    fprintf(stderr, "\t-> Ignored (not a valid source host)\n");
+	    Perror("\t-> Ignored (not a valid source host)\n");
 	
 	return NULL;
     }
@@ -580,7 +636,7 @@ ypresp_key_val *ypproc_first_2_svc(ypreq_nokey *key,
 	free(result.key.keydat_val);
 	result.key.keydat_val = NULL;
     }
-    
+ 
     result.val.valdat_len = 0;
     if (result.val.valdat_val)
     {
@@ -619,14 +675,14 @@ ypresp_key_val *ypproc_first_2_svc(ypreq_nokey *key,
     if (debug_flag)
     {
 	if (result.stat == YP_TRUE)
-	    fprintf(stderr, "\t-> Key = \"%.*s\", Value = \"%.*s\"\n",
+	    Perror("\t-> Key = \"%.*s\", Value = \"%.*s\"\n",
 		    (int) result.key.keydat_len,
 		    result.key.keydat_val,
 		    (int) result.val.valdat_len,
 		    result.val.valdat_val);
 		    
 	else
-	    fprintf(stderr, "\t-> Error #%d\n", result.stat);
+	    Perror("\t-> Error #%d\n", result.stat);
     }
     
     return &result;
@@ -644,13 +700,13 @@ ypresp_key_val *ypproc_next_2_svc(ypreq_key *key,
 
     if (debug_flag)
     {
-	fprintf(stderr, "ypproc_next(): [From: %s:%d]\n",
+	Perror("ypproc_next(): [From: %s:%d]\n",
 		inet_ntoa(rqhost->sin_addr),
 		ntohs(rqhost->sin_port));
 		
-	fprintf(stderr, "\tdomainname = \"%s\"\n", key->domain);
-	fprintf(stderr, "\tmapname = \"%s\"\n", key->map);
-	fprintf(stderr, "\tkeydat = \"%.*s\"\n",
+	Perror("\tdomainname = \"%s\"\n", key->domain);
+	Perror("\tmapname = \"%s\"\n", key->map);
+	Perror("\tkeydat = \"%.*s\"\n",
 		(int) key->key.keydat_len,
 		key->key.keydat_val);
     }
@@ -658,7 +714,7 @@ ypresp_key_val *ypproc_next_2_svc(ypreq_key *key,
     if (!is_valid_host(rqhost))
     {
 	if (debug_flag)
-	    fprintf(stderr, "\t-> Ignored (not a valid source host)\n");
+	    Perror("\t-> Ignored (not a valid source host)\n");
 	
 	return NULL;
     }
@@ -669,7 +725,7 @@ ypresp_key_val *ypproc_next_2_svc(ypreq_key *key,
 	free(result.key.keydat_val);
 	result.key.keydat_val = NULL;
     }
-    
+
     result.val.valdat_len = 0;
     if (result.val.valdat_val)
     {
@@ -711,13 +767,13 @@ ypresp_key_val *ypproc_next_2_svc(ypreq_key *key,
     if (debug_flag)
     {
 	if (result.stat == YP_TRUE)
-	    fprintf(stderr, "\t-> Key = \"%.*s\", Value = \"%.*s\"\n",
+	    Perror("\t-> Key = \"%.*s\", Value = \"%.*s\"\n",
 		    (int) result.key.keydat_len,
 		    result.key.keydat_val,
 		    (int) result.val.valdat_len,
 		    result.val.valdat_val);
 	else
-	    fprintf(stderr, "\t-> Error #%d\n", result.stat);
+	    Perror("\t-> Error #%d\n", result.stat);
     }
     
     return &result;
@@ -727,49 +783,85 @@ ypresp_key_val *ypproc_next_2_svc(ypreq_key *key,
 
 static void print_ypmap_parms(const struct ypmap_parms *pp)
 {
-    fprintf(stderr, "\t\tdomain   = \"%s\"\n", pp->domain);
-    fprintf(stderr, "\t\tmap      = \"%s\"\n", pp->map);
-    fprintf(stderr, "\t\tordernum = %u\n", pp->ordernum);
-    fprintf(stderr, "\t\tpeer     = \"%s\"\n", pp->peer);
+    Perror("\t\tdomain   = \"%s\"\n", pp->domain);
+    Perror("\t\tmap      = \"%s\"\n", pp->map);
+    Perror("\t\tordernum = %u\n", pp->ordernum);
+    Perror("\t\tpeer     = \"%s\"\n", pp->peer);
 }
 
 
 /*
-** XXXX Implement Me!
+** Clean up after child processes signal their termination.
+*/
+void reapchild(int sig)
+{
+   int st;
+
+#ifdef HAVE_WAIT3   
+   while (wait3(&st, WNOHANG, NULL) > 0)
+#else 
+   while (waitpid(-1, &st, WNOHANG) > 0)
+#endif
+      children--;
+}
+
+/*
+** Stole the ypxfr implementation from the yps package.
 */
 ypresp_xfr *ypproc_xfr_2_svc(ypreq_xfr *xfr,
 			     struct svc_req *rqstp)
 {
     static ypresp_xfr result;
     struct sockaddr_in *rqhost;
-
+    char ypxfr_command[MAXPATHLEN];
 
     rqhost = svc_getcaller(rqstp->rq_xprt);
 
     if (debug_flag)
     {
-	fprintf(stderr, "ypproc_xfr_2_svc(): [From: %s:%d]\n\tmap_parms:\n",
+	Perror("ypproc_xfr_2_svc(): [From: %s:%d]\n\tmap_parms:\n",
 		inet_ntoa(rqhost->sin_addr),
 		ntohs(rqhost->sin_port));
 	
 	print_ypmap_parms(&xfr->map_parms);
-	fprintf(stderr, "\t\ttransid = %u\n", xfr->transid);
-	fprintf(stderr, "\t\tprog = %u\n", xfr->prog);
-	fprintf(stderr, "\t\tport = %u\n", xfr->port);
-	fprintf(stderr, "\t-> Request denied.\n");
+	Perror("\t\ttransid = %u\n", xfr->transid);
+	Perror("\t\tprog = %u\n", xfr->prog);
+	Perror("\t\tport = %u\n", xfr->port);
     }
 
     if (!is_valid_host(rqhost))
     {
 	if (debug_flag)
-	    fprintf(stderr, "\t-> Ignored (not a valid source host)\n");
-	
+	    Perror("\t-> Ignored (not a valid source host)\n");
+
 	return NULL;
     }
 
+    switch(fork())
+    {
+       case 0:
+           {
+              char g[11], t[11], p[11];
+              
+              sprintf( ypxfr_command, "%s/ypxfr", YPBINDIR);
+              sprintf( t, "%u", xfr->transid);
+              sprintf( g, "%u", xfr->prog);
+              sprintf( p, "%u", xfr->port);
+              execl(ypxfr_command, "ypxfr", "-d", xfr->map_parms.domain, "-h",
+                  xfr->map_parms.peer, "-f", "-C", t, g,
+                  inet_ntoa(rqhost->sin_addr), p, xfr->map_parms.map, NULL);
+              Perror("ypxfr execl(): %s", strerror(errno));
+              exit(0);
+           }
+       case -1:
+           Perror("fork(): %s",strerror(errno));
+           result.xfrstat = YPXFR_XFRERR;
+       default:
+           result.xfrstat = YPXFR_SUCC;
+           break;
+    }
+
     result.transid = xfr->transid;
-    result.xfrstat = YPXFR_REFUSED;
-    
     return &result;
 }
 
@@ -784,14 +876,14 @@ void *ypproc_clear_2_svc(void *dummy,
     rqhost = svc_getcaller(rqstp->rq_xprt);
 
     if (debug_flag)
-    	fprintf(stderr, "ypproc_clear_2_svc() [From: %s:%d]\n",
+    	Perror("ypproc_clear_2_svc() [From: %s:%d]\n",
 		inet_ntoa(rqhost->sin_addr),
 		ntohs(rqhost->sin_port));
 		
     if (!is_valid_host(rqhost))
     {
 	if (debug_flag)
-	    fprintf(stderr, "\t-> Ignored (not a valid source host)\n");
+	    Perror("\t-> Ignored (not a valid source host)\n");
 	
 	return NULL;
     }
@@ -804,7 +896,7 @@ static int ypall_close(void *data)
 {
     if (debug_flag && data == NULL)
     {
-	fprintf(stderr, "ypall_close() called with NULL pointer.\n");
+	Perror("ypall_close() called with NULL pointer.\n");
 	return 0;
     }
 	    
@@ -845,31 +937,38 @@ ypresp_all *ypproc_all_2_svc(ypreq_nokey *nokey,
     extern xdr_ypall_cb_t xdr_ypall_cb;
     struct sockaddr_in *rqhost;
 
-    
     rqhost = svc_getcaller(rqstp->rq_xprt);
 
     if (debug_flag)
     {
-	fprintf(stderr, "ypproc_all_2_svc(): [From: %s:%d]\n",
+	Perror("ypproc_all_2_svc(): [From: %s:%d]\n",
 		inet_ntoa(rqhost->sin_addr),
 		ntohs(rqhost->sin_port));
 		
-	fprintf(stderr, "\t\tdomain = \"%s\"\n", nokey->domain);
-	fprintf(stderr, "\t\tmap = \"%s\"\n", nokey->map);
+	Perror("\t\tdomain = \"%s\"\n", nokey->domain);
+	Perror("\t\tmap = \"%s\"\n", nokey->map);
     }
 
     if (!is_valid_host(rqhost))
     {
 	if (debug_flag)
-	    fprintf(stderr, "\t-> Ignored (not a valid source host)\n");
+	    Perror("\t-> Ignored (not a valid source host)\n");
 	
 	return NULL;
     }
 
+    if(children < MAX_CHILDREN && fork())
+    {
+       children++;
+       return NULL;
+    }
+    else
+      forked++;
+
     xdr_ypall_cb.u.encode = NULL;
     xdr_ypall_cb.u.close  = NULL;
     xdr_ypall_cb.data = NULL;
-    
+
     result.more = TRUE;
     
     if (nokey->map[0] == '\0' || nokey->domain[0] == '\0')
@@ -904,7 +1003,7 @@ ypresp_all *ypproc_all_2_svc(ypreq_nokey *nokey,
 		xdr_ypall_cb.data = (void *) dbp;
 
 		if (debug_flag)
-		    fprintf(stderr, "\t -> First value returned.\n");
+		    Perror("\t -> First value returned.\n");
 			
 		return &result;
 	    }
@@ -914,7 +1013,7 @@ ypresp_all *ypproc_all_2_svc(ypreq_nokey *nokey,
     }
     
     if (debug_flag)
-	fprintf(stderr, "\t -> Error should have been returned.\n");
+	Perror("\t -> Error should have been returned.\n");
 	    
     return &result;
 }
@@ -923,26 +1022,27 @@ ypresp_all *ypproc_all_2_svc(ypreq_nokey *nokey,
 ypresp_master *ypproc_master_2_svc(ypreq_nokey *nokey,
 				   struct svc_req *rqstp)
 {
+    static char hostbuf[1025];
     static ypresp_master result;
     struct sockaddr_in *rqhost;
-    
-    
+
+
     rqhost = svc_getcaller(rqstp->rq_xprt);
 
     if (debug_flag)
     {
-	fprintf(stderr, "ypproc_master_2_svc(): [From: %s:%d]\n",
+	Perror("ypproc_master_2_svc(): [From: %s:%d]\n",
 		inet_ntoa(rqhost->sin_addr),
 		ntohs(rqhost->sin_port));
 		
-	fprintf(stderr, "\t\tdomain = \"%s\"\n", nokey->domain);
-	fprintf(stderr, "\t\tmap = \"%s\"\n", nokey->map);
+	Perror("\t\tdomain = \"%s\"\n", nokey->domain);
+	Perror("\t\tmap = \"%s\"\n", nokey->map);
     }
 
     if (!is_valid_host(rqhost))
     {
 	if (debug_flag)
-	    fprintf(stderr, "\t-> Ignored (not a valid source host)\n");
+	    Perror("\t-> Ignored (not a valid source host)\n");
 	
 	return NULL;
     }
@@ -973,13 +1073,20 @@ ypresp_master *ypproc_master_2_svc(ypreq_nokey *nokey,
 	    if (val.dptr == NULL)
 	    {
 		/* No YP_MASTER_NAME record in map? Assume we are Master */
-		static char hostbuf[1025];
-
 		gethostname(hostbuf, sizeof(hostbuf)-1);
 		result.peer = strdup(hostbuf);
 	    }
 	    else
-		result.peer = val.dptr;
+	    {
+	        int i;
+
+		/* put the eof string mark at the end of the string */
+	        for(i=0; i<val.dsize; i++)
+		   hostbuf[i] = val.dptr[i];
+	        hostbuf[val.dsize] = '\0';
+
+ 		result.peer = strdup(hostbuf);
+	    }
 
 	    result.stat = YP_TRUE;
 	    gdbm_close(dbp);
@@ -990,7 +1097,7 @@ ypresp_master *ypproc_master_2_svc(ypreq_nokey *nokey,
 	result.peer = strdup("");
 
     if (debug_flag)
-	fprintf(stderr, "\t-> Peer = \"%s\"\n", result.peer);
+      Perror("\t-> Peer = \"%s\"\n", result.peer);
     
     return &result;
 }
@@ -1007,18 +1114,18 @@ ypresp_order *ypproc_order_2_svc(ypreq_nokey *nokey,
 
     if (debug_flag)
     {
-	fprintf(stderr, "ypproc_order_2_svc(): [From: %s:%d]\n",
+	Perror("ypproc_order_2_svc(): [From: %s:%d]\n",
 		inet_ntoa(rqhost->sin_addr),
 		ntohs(rqhost->sin_port));
 		
-	fprintf(stderr, "\t\tdomain = \"%s\"\n", nokey->domain);
-	fprintf(stderr, "\t\tmap = \"%s\"\n", nokey->map);
+	Perror("\t\tdomain = \"%s\"\n", nokey->domain);
+	Perror("\t\tmap = \"%s\"\n", nokey->map);
     }
 
     if (!is_valid_host(rqhost))
     {
 	if (debug_flag)
-	    fprintf(stderr, "\t-> Ignored (not a valid source host)\n");
+	    Perror("\t-> Ignored (not a valid source host)\n");
 	
 	return NULL;
     }
@@ -1059,7 +1166,7 @@ ypresp_order *ypproc_order_2_svc(ypreq_nokey *nokey,
     }
 
     if (debug_flag)
-	fprintf(stderr, "-> Order # %d\n", result.ordernum);
+	Perror("-> Order # %d\n", result.ordernum);
     
     return &result;
 }
@@ -1112,17 +1219,17 @@ ypresp_maplist *ypproc_maplist_2_svc(domainname *name,
 
     if (debug_flag)
     {
-	fprintf(stderr, "ypproc_maplist_2_svc(): [From: %s:%d]\n",
+	Perror("ypproc_maplist_2_svc(): [From: %s:%d]\n",
 		inet_ntoa(rqhost->sin_addr),
 		ntohs(rqhost->sin_port));
 		
-	fprintf(stderr, "\t\tdomain = \"%s\"\n", *name);
+	Perror("\t\tdomain = \"%s\"\n", *name);
     }
 
     if (!is_valid_host(rqhost))
     {
 	if (debug_flag)
-	    fprintf(stderr, "\t-> Ignored (not a valid source host)\n");
+	    Perror("\t-> Ignored (not a valid source host)\n");
 	
 	return NULL;
     }
@@ -1132,8 +1239,6 @@ ypresp_maplist *ypproc_maplist_2_svc(domainname *name,
     
     result.maps = NULL;
 
-    result.stat = YP_TRUE;
-
     if ((*name)[0] == '\0')
 	result.stat = YP_BADARGS;
     else if (!is_valid_domain(*name))
@@ -1141,16 +1246,14 @@ ypresp_maplist *ypproc_maplist_2_svc(domainname *name,
     else
     {
 	DIR *dp;
-
-	fprintf( stderr, "Reading dir %s\n", *name);
+        
 	/* open domain directory */
 	dp = opendir( *name);
 	if (dp == NULL)
 	{
 	    if (debug_flag)
 	    {
-		fprintf(stderr, "%s: opendir: ", progname);
-		perror(".");
+		Perror("%s: opendir: %s", progname, strerror(errno));
 	    }
 	    
 	    result.stat = YP_BADDB;
@@ -1159,11 +1262,12 @@ ypresp_maplist *ypproc_maplist_2_svc(domainname *name,
 	{
 	    struct dirent *dep;
 	    
-	    while ((dep = readdir(dp)) != NULL) {
+	    while ((dep = readdir(dp)) != NULL) 
+	    {
 	        /* ignore files . and .. */
 	        if (strcmp( ".", dep->d_name) == 0 ||
-		    strcmp( "..", dep->d_name) == 0)
-		  continue;
+	            strcmp( "..", dep->d_name) == 0)
+                  continue;
 		if (add_maplist(&result.maps, dep->d_name) < 0)
 		{
 		    result.stat = YP_YPERR;
@@ -1171,6 +1275,7 @@ ypresp_maplist *ypproc_maplist_2_svc(domainname *name,
 		}
 	    }
 	    closedir(dp);
+	    result.stat = YP_TRUE;
 	}
     }
     
@@ -1181,19 +1286,19 @@ ypresp_maplist *ypproc_maplist_2_svc(domainname *name,
 	    ypmaplist *p;
 
 	    p = result.maps;
-	    fprintf(stderr, "-> ");
+	    Perror("-> ");
 	    while (p)
 	    {
-		fprintf(stderr, "%s", p->map);
+		Perror("%s", p->map);
 		if (p->next) {
-		    fprintf(stderr, ", ");
+		    Perror(", ");
 		}
 		p = p->next;
 	    }
 	    putc('\n', stderr);
 	}
 	else
-	    fprintf(stderr, "\t-> Error #%d\n", result.stat);
+	    Perror("\t-> Error #%d\n", result.stat);
     }
     
     return &result;
